@@ -2,7 +2,7 @@
 # -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
 #   Catfish - a versatile file searching tool
 #   Copyright (C) 2007-2012 Christian Dywan <christian@twotoasts.de>
-#   Copyright (C) 2012-2019 Sean Davis <bluesabre@xfce.org>
+#   Copyright (C) 2012-2020 Sean Davis <bluesabre@xfce.org>
 #
 #   This program is free software: you can redistribute it and/or modify it
 #   under the terms of the GNU General Public License version 2, as published
@@ -16,6 +16,10 @@
 #   You should have received a copy of the GNU General Public License along
 #   with this program.  If not, see <https://www.gnu.org/licenses/>.
 
+# pylint: disable=C0413
+# pylint: disable=C0114
+# pylint: disable=C0116
+
 import logging
 
 import io
@@ -26,33 +30,34 @@ import subprocess
 import time
 from itertools import permutations
 
-from mimetypes import guess_type
-
-from sys import version_info
+import mimetypes
 
 import gi
 gi.require_version('GLib', '2.0')  # noqa
 from gi.repository import GLib
 
 try:
-    from zeitgeist.client import ZeitgeistDBusInterface
-    from zeitgeist.datamodel import Event, TimeRange
-    from zeitgeist import datamodel
-    iface = ZeitgeistDBusInterface()
-    zeitgeist_support = True
-except Exception:
-    zeitgeist_support = False
+    gi.require_version('Zeitgeist', '2.0')
+    from gi.repository import Zeitgeist
+    log_trial = Zeitgeist.Log.get_default()
+    if log_trial.datapath() is None:
+        ZEITGEIST_SUPPORT = False
+    else:
+        ZEITGEIST_SUPPORT = True
+except ImportError:
+    ZEITGEIST_SUPPORT = False
+except ValueError:
+    ZEITGEIST_SUPPORT = False
 
-logger = logging.getLogger('catfish_search')
-python3 = version_info[0] > 2
-engine_count = 0
+LOGGER = logging.getLogger('catfish_search')
+ENGINE_COUNT = 0
 
 FNULL = open(os.devnull, 'w')
-if subprocess.call(['which', 'locate'],
+if subprocess.call(['which', 'locate'],  # pylint: disable=R1703
                    stdout=FNULL, stderr=subprocess.STDOUT) == 0:
-    locate_support = True
+    LOCATE_SUPPORT = True
 else:
-    locate_support = False
+    LOCATE_SUPPORT = False
 FNULL.close()
 
 
@@ -69,7 +74,7 @@ def get_keyword_list(keywords):
     return kwords
 
 
-def string_regex(keywords, path):
+def string_regex(keywords, path):  # pylint: disable=W0613
     """Returns a string with the regular expression containing all combinations
     of the keywords."""
     if len(keywords) == 0:
@@ -108,18 +113,21 @@ class CatfishSearchEngine:
          walk             'os.walk' to search for files (like find).
          zeitgeist        Zeitgeist indexing service to search for files.
         """
-        global engine_count
-        engine_count += 1
-        self.engine_id = engine_count
-        logger.debug(
+        self.stop_time = 0
+        self.keywords = ""
+
+        global ENGINE_COUNT
+        ENGINE_COUNT += 1
+        self.engine_id = ENGINE_COUNT
+        LOGGER.debug(
             "[%i] engine initializing with methods: %s",
             self.engine_id, str(methods))
         self.methods = []
         if 'zeitgeist' in methods:
-            if zeitgeist_support:
+            if ZEITGEIST_SUPPORT:
                 self.add_method(CatfishSearchMethod_Zeitgeist)
         if 'locate' in methods:
-            if locate_support:
+            if LOCATE_SUPPORT:
                 self.add_method(CatfishSearchMethod_Locate)
         if 'fulltext' in methods:
             self.add_method(CatfishSearchMethod_Fulltext)
@@ -129,19 +137,19 @@ class CatfishSearchEngine:
         initialized = []
         for method in self.methods:
             initialized.append(method.method_name)
-        logger.debug(
+        LOGGER.debug(
             "[%i] engine initialized with methods: %s",
             self.engine_id, str(initialized))
         self.start_time = 0.0
 
     def __del__(self):
-        logger.debug("[%i] engine destroyed", self.engine_id)
+        LOGGER.debug("[%i] engine destroyed", self.engine_id)
 
     def add_method(self, method_class):
         """Add a CatfishSearchMethod the the engine's search backends."""
         self.methods.append(method_class())
 
-    def run(self, keywords, path, limit=-1, regex=False): # noqa
+    def run(self, keywords, path, limit=-1, regex=False):  # noqa
         """Run the CatfishSearchEngine.
 
         Each method is run sequentially in the order they are added.
@@ -155,7 +163,7 @@ class CatfishSearchEngine:
         keywords = get_keyword_list(keywords)
         self.keywords = " ".join(keywords)
 
-        logger.debug("[%i] path: %s, keywords: %s, limit: %i, regex: %s",
+        LOGGER.debug("[%i] path: %s, keywords: %s, limit: %i, regex: %s",
                      self.engine_id, str(path), str(keywords), limit,
                      str(regex))
 
@@ -167,6 +175,11 @@ class CatfishSearchEngine:
             else:
                 keys.append(key)
 
+        # path may be empty when some "extension" scheme is used in URL
+        if not path:
+            self.stop()
+            return
+
         # For simplicity, make sure the path contains a trailing '/'
         if not path.endswith('/'):
             path += '/'
@@ -181,9 +194,9 @@ class CatfishSearchEngine:
         file_count = 0
         for method in self.methods:
             if self.stop_time > 0:
-                logger.debug("Engine is stopped")
+                LOGGER.debug("Engine is stopped")
                 return
-            logger.debug(
+            LOGGER.debug(
                 "[%i] Starting search method: %s",
                 self.engine_id, method.method_name)
             for filename in method.run(keywords, path, regex,
@@ -193,7 +206,7 @@ class CatfishSearchEngine:
                     for filepath in exclude:
                         if filename.startswith(filepath):
                             if self.stop_time > 0:
-                                logger.debug("Engine is stopped")
+                                LOGGER.debug("Engine is stopped")
                                 return
                             found_bad = True
                     if found_bad:
@@ -261,7 +274,7 @@ class CatfishSearchEngine:
             method.stop()
         self.stop_time = time.time()
         clock = self.stop_time - self.start_time
-        logger.debug("[%i] Last query: %f seconds", self.engine_id, clock)
+        LOGGER.debug("[%i] Last query: %f seconds", self.engine_id, clock)
 
 
 class CatfishSearchMethod:
@@ -273,7 +286,7 @@ class CatfishSearchMethod:
         """Base CatfishSearchMethod Initializer."""
         self.method_name = method_name
 
-    def run(self, keywords, path, regex=False, exclude_paths=[]):
+    def run(self, keywords, path, regex=False, exclude_paths=[]):  # pylint: disable=W0613
         """Base CatfishSearchMethod run method."""
         return NotImplemented
 
@@ -295,8 +308,10 @@ class CatfishSearchMethod_Walk(CatfishSearchMethod):
     def __init__(self):
         """Initialize the 'walk' Search Method."""
         CatfishSearchMethod.__init__(self, "walk")
+        self.running = False
 
-    def get_dir_list(self, root, dirs, xdg_list, exclude_list):
+    def get_dir_list(self, root, dirs, xdg_list,
+                     exclude_list, processed_links):
         dirs = sorted(dirs, key=lambda s: s.lower())
 
         # Prioritize: XDG, Visible (Linked), Dotfile (Linked)
@@ -305,16 +320,32 @@ class CatfishSearchMethod_Walk(CatfishSearchMethod):
         dotlinks = []
         notdotdirs = []
         notdotlinks = []
+
+        realroot = os.path.realpath(root)
+
         for path in dirs:
             path = os.path.join(root, path)
-            if not path.endswith("/"):
-                path = path + "/"
+            # Remove trailing slashes to ensure that calling os.path.basename()
+            # will not cut the path to an empty string
+            path = os.path.normpath(path)
             if path in exclude_list:
                 continue
             if path in xdg_list:
                 xdgdirs.append(path)
                 continue
             islink = os.path.islink(path)
+            if islink:
+                realpath = os.path.realpath(path)
+                if realpath in processed_links:
+                    continue
+                # Respect the user's settings
+                if realpath in exclude_list:
+                    continue
+                # Sandbox search results to the starting path, don't allow:
+                # Start: /home/username/
+                # Link: ~/.wine/dosdevices/z:/ -> /
+                if realroot.startswith(realpath):
+                    continue
             if os.path.basename(path).startswith("."):
                 if islink:
                     dotlinks.append(path)
@@ -329,17 +360,6 @@ class CatfishSearchMethod_Walk(CatfishSearchMethod):
         dirlist = xdgdirs + notdotdirs + notdotlinks + dotdirs + dotlinks
         return dirlist
 
-    def get_root_list(self, path, xdg_list, exclude_list):
-        # Sort the roots alphabetically
-        roots = [d for d in os.listdir(path)
-                 if os.path.isdir(os.path.join(path, d)) and
-                 os.path.join(path, d) not in exclude_list]
-        roots = sorted(roots, key=lambda s: s.lower())
-        results = []
-        for dirpath in self.get_dir_list(path, roots, xdg_list, exclude_list):
-            results.append(os.path.join(path, dirpath))
-        return results
-
     def run(self, keywords, path, regex=False, exclude_paths=[]):
         """Run the search method using keywords and path.  regex is not used
         by this search method.
@@ -361,42 +381,42 @@ class CatfishSearchMethod_Walk(CatfishSearchMethod):
 
         # Grab the special directory list to get them precedence
         xdgdirlist = [
-            GLib.get_user_special_dir(GLib.USER_DIRECTORY_DESKTOP),
-            GLib.get_user_special_dir(GLib.USER_DIRECTORY_DOCUMENTS),
-            GLib.get_user_special_dir(GLib.USER_DIRECTORY_DOWNLOAD),
-            GLib.get_user_special_dir(GLib.USER_DIRECTORY_MUSIC),
-            GLib.get_user_special_dir(GLib.USER_DIRECTORY_PICTURES),
-            GLib.get_user_special_dir(
-                GLib.USER_DIRECTORY_PUBLIC_SHARE),
-            GLib.get_user_special_dir(GLib.USER_DIRECTORY_TEMPLATES),
-            GLib.get_user_special_dir(GLib.USER_DIRECTORY_VIDEOS),
+            GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_DESKTOP),
+            GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_DOCUMENTS),
+            GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_DOWNLOAD),
+            GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_MUSIC),
+            GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_PICTURES),
+            GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_PUBLIC_SHARE),
+            GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_TEMPLATES),
+            GLib.get_user_special_dir(GLib.UserDirectory.DIRECTORY_VIDEOS),
         ]
 
-        for path in self.get_root_list(path, xdgdirlist, exclude):
-            for root, dirs, files in os.walk(top=path, topdown=True,
-                                             onerror=None,
-                                             followlinks=True):
-                # Bail once the search has been canceled
-                if not self.running:
-                    break
+        for root, dirs, files in os.walk(top=path, topdown=True,
+                                         onerror=None,
+                                         followlinks=True):
+            # Bail once the search has been canceled
+            if not self.running:
+                break
 
-                # Check if we've already processed symbolic paths
-                if os.path.islink(root):
-                    realpath = os.path.realpath(root)
-                    if realpath in processed_links:
-                        yield True
-                        continue
-                    processed_links.append(realpath)
+            # Check if we've already processed symbolic paths
+            if os.path.islink(root):
+                realpath = os.path.realpath(root)
+                if realpath in processed_links:
+                    yield True
+                    continue
+                processed_links.append(realpath)
 
-                # Prioritize and drop excluded paths
-                dirs[:] = self.get_dir_list(root, dirs, xdgdirlist, exclude)
+            # Prioritize and drop excluded paths
+            dirs[:] = self.get_dir_list(
+                root, dirs, xdgdirlist, exclude, processed_links)
 
-                paths = dirs + files
-                paths.sort()
-                for path in paths:
-                    if any(keyword in path.lower() for keyword in keywords):
-                        yield os.path.join(root, path)
-                yield True
+            paths = sorted(dirs + files)
+
+            # Check paths in the second and deeper levels of the selected
+            # directory
+            for path in paths:
+                if any(keyword in path.lower() for keyword in keywords):
+                    yield os.path.join(root, path)
             yield True
         yield False
 
@@ -422,6 +442,56 @@ class CatfishSearchMethod_Fulltext(CatfishSearchMethod):
         self.running = False
         self.exact = False
 
+    def check_charset(self, root, filename):
+        """Checks file for BOM to detect UTF encodings.
+        Also checks file for null byte. Most files containing
+        a null byte will be binary. UTF-16 and UTF-32 without
+        BOM can contain null bytes, support for that can be
+        added by checking the line ending."""
+        file = open(os.path.join(root, filename), 'rb')
+        try:
+            data = file.read(64).lower()
+
+            if data.startswith(b'\x00\x00\xfe\xff'):
+                return 'utf-32-be'
+
+            elif data.startswith(b'\xff\xfe\x00\x00'):
+                return 'utf-32-le'
+
+            elif data.startswith(b'\x2b\x2f\x76'):
+                return 'utf-7'
+
+            elif data.startswith(b'\xfe\xff'):
+                return 'utf-16-be'
+
+            elif data.startswith(b'\xff\xfe'):
+                return 'utf-16-le'
+
+            elif b'\0' in data:
+                return 'binary'
+
+            else:
+                return 'utf-8'
+
+        finally:
+            file.close()
+
+    def is_txt(self, filename):
+        """Checks if text mimetype."""
+        mime = str(mimetypes.guess_type(filename)[0])
+        text_list = ('ardour', 'audacity', 'desktop', 'document',
+                     'fontforge', 'java', 'json', 'm4', 'mbox',
+                     'mimearchive', 'msg', 'none', 'perl',
+                     'pgp-keys', 'php', 'postscript', 'rtf',
+                     'ruby', 'shellscript', 'spreadsheet', 'sql',
+                     'subrip', 'text', 'url', 'winhlp',
+                     'x-bittorent', 'x-cue', 'x-extension-cfg',
+                     'x-glade', 'x-mpegurl', 'x-sami', 'x-theme',
+                     'x-trash', 'xml', 'xpixmap', 'yaml')
+        for filetype in text_list:
+            if filetype in mime.lower():
+                return True
+
     def run(self, keywords, path, regex=False, exclude_paths=[]):  # noqa
         """Run the search method using keywords and path.  regex is not used
         by this search method.
@@ -441,53 +511,60 @@ class CatfishSearchMethod_Fulltext(CatfishSearchMethod):
                     find_keywords_backup.append(keyword)
 
         # Start walking the folder structure.
-        for root, dirs, files in os.walk(path):
+        for root, dirs, files in os.walk(path):  # pylint: disable=W0612
             if self.force_stop:
                 break
+            # Don't search user excluded directories.
+            if root.startswith(tuple(exclude_paths)):
+                dirs[:] = []
+                continue
 
             for filename in files:
-                if self.force_stop:
-                    break
+                try:
+                    fullpath = os.path.join(root, filename)
 
-                # If the filetype is known to not be text, move along.
-                mime = guess_type(filename)[0]
-                if not mime or 'text' in mime:
-                    try:
-                        opened = open(os.path.join(root, filename), 'r')
+                    # Skip if special file.
+                    if not os.path.isfile(fullpath):
+                        continue
+                    if os.path.getsize(fullpath) == 0:
+                        continue
+                    # Skip if not text file.
+                    if not self.is_txt(filename):
+                        continue
+                    # Check character encoding, skip if binary.
+                    charset = self.check_charset(root, filename)
+                    if charset == 'binary':
+                        continue
 
-                        find_keywords = find_keywords_backup
-
-                        # Check each line.  If a keyword is found, yield.
-                        try:
-                            for line in opened:
-                                if self.force_stop:
+                    # Check each line. If a keyword is found, yield.
+                    open_file = open(fullpath, 'r', encoding=charset)
+                    with open_file as file_text:
+                        if self.exact:
+                            for line in file_text:
+                                if " ".join(keywords) in line:
+                                    yield fullpath
+                        else:
+                            match_list = set()
+                            for line in file_text:
+                                for kword in keywords:
+                                    if kword in line.lower():
+                                        match_list.add(kword)
+                                if len(set(keywords)) == len(match_list):
                                     break
-
-                                if self.exact:
-                                    if " ".join(keywords) in line:
-                                        yield os.path.join(root, filename)
-                                        break
-                                else:
-                                    if any(keyword in line.lower()
-                                            for keyword in keywords):
-                                        found_keywords = []
-                                        for find_keyword in find_keywords:
-                                            if find_keyword in line.lower():
-                                                found_keywords.append(
-                                                    find_keyword)
-                                        for found_keyword in found_keywords:
-                                            find_keywords.remove(found_keyword)
-
-                                        if len(find_keywords) == 0:
-                                            yield os.path.join(root, filename)
-                                            break
-                        except UnicodeDecodeError:
-                            pass
-
-                        opened.close()
-                    except IOError:
-                        pass
-                yield True
+                            if len(set(keywords)) == len(match_list):
+                                yield fullpath
+                # Skips on errors, move on to next in list.
+                except UnicodeDecodeError:
+                    continue
+                except UnicodeError:
+                    continue
+                except FileNotFoundError:
+                    continue
+                except PermissionError:
+                    continue
+                except OSError:
+                    continue
+            yield True
         yield False
         self.force_stop = False
         self.running = False
@@ -509,30 +586,48 @@ class CatfishSearchMethod_Zeitgeist(CatfishSearchMethod):
     def __init__(self):
         """Initialize the Zeitgeist SearchMethod."""
         CatfishSearchMethod.__init__(self, "zeitgeist")
+        self.stop_search = False
+        self.events = []
+
+    def result_callback(self, log, result, data):
+        events = log.find_events_finish(result)
+        for i in range(events.size()):  # cannot iterate events directly
+            event = events.next_value()
+            if event is None:
+                continue
+            self.events.append(event)
+        self.stop_search = True
 
     def run(self, keywords, path, regex=False, exclude_paths=[]):
         """Run the Zeitgeist SearchMethod."""
+        keywords = " ".join(keywords)
         self.stop_search = False
-        event_template = Event()
-        time_range = TimeRange.from_seconds_ago(60 * 3600 * 24)
-        # 60 days at most
+        event_template = Zeitgeist.Event()
 
-        results = iface.FindEvents(
-            time_range,  # (min_timestamp, max_timestamp) in milliseconds
+        # 60 days at most
+        end = int(time.time() * 1000)
+        start = end - (60 * 3600 * 24 * 1000)
+        time_range = Zeitgeist.TimeRange.new(start, end)
+
+        self.events = []
+        log = Zeitgeist.Log.get_default()
+        log.find_events(
+            time_range,
             [event_template, ],
-            datamodel.StorageState.Any,
+            Zeitgeist.StorageState.ANY,
             1000,
-            datamodel.ResultType.MostRecentSubjects
+            Zeitgeist.ResultType.MOST_RECENT_SUBJECTS,
+            None, self.result_callback, None
         )
 
-        results = (datamodel.Event(result) for result in results)
+        while self.stop_search == False:
+            yield False
+
         uniques = []
 
-        for event in results:
-            if self.stop_search:
-                break
+        for event in self.events:
             for subject in event.get_subjects():
-                uri = str(subject.uri)
+                uri = subject.get_uri()
                 if uri.startswith('file://'):
                     fullname = str(uri[7:])
                     filepath, filename = os.path.split(fullname)
@@ -541,7 +636,6 @@ class CatfishSearchMethod_Zeitgeist(CatfishSearchMethod):
                             path in filepath:
                         uniques.append(uri)
                         yield fullname
-        self.stop_search = True
 
     def stop(self):
         """Stop the Zeitgeist SearchMethod."""
@@ -564,7 +658,7 @@ class CatfishSearchMethodExternal(CatfishSearchMethod):
         self.command = []
         self.process = None
 
-    def assemble_query(self, keywords, path):
+    def assemble_query(self, keywords, path):  # pylint: disable=W0613
         """Base assemble_query method."""
         return False
 
@@ -593,8 +687,7 @@ class CatfishSearchMethodExternal(CatfishSearchMethod):
             return map(lambda s: s.decode(encoding='UTF8',
                                           errors='replace').strip(),
                        output.readlines())
-        else:
-            return output
+        return output
 
     def status(self):
         """Return the current search status."""
@@ -638,7 +731,7 @@ class CatfishSearchMethod_Locate(CatfishSearchMethodExternal):
             "regex": False
         }
         try:
-            details = subprocess.check_output(["locate", "--help"])
+            details = subprocess.check_output(["locate", "--help"],stderr=subprocess.DEVNULL)
             details = details.decode("utf-8")
             if "--existing" in details:
                 caps["existing"] = True
@@ -654,5 +747,4 @@ class CatfishSearchMethod_Locate(CatfishSearchMethodExternal):
         if self.caps["regex"]:
             return ["locate", "--regex", "--basename", "-i",
                     "{}".format(string_regex(keywords, path))]
-        else:
-            return ["locate", "-i", "%path*", str(keywords)]
+        return ["locate", "-i", "%path*", str(keywords)]
